package sun.security.provider.certpath;

import java.security.AccessController;
import sun.security.action.GetBooleanAction;
import java.util.Iterator;
import java.security.cert.Certificate;
import java.util.HashSet;
import java.util.Collections;
import sun.security.x509.GeneralSubtrees;
import sun.security.x509.GeneralNames;
import sun.security.x509.SubjectAlternativeNameExtension;
import sun.security.x509.X509CertImpl;
import sun.security.x509.NameConstraintsExtension;
import sun.security.x509.X500Name;
import sun.security.x509.GeneralNameInterface;
import java.util.LinkedList;
import java.security.GeneralSecurityException;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.CertStoreException;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.security.cert.CertStore;
import java.util.List;
import java.security.cert.X509CertSelector;
import java.util.Set;
import sun.security.util.Debug;

public abstract class Builder
{
    private static final Debug debug;
    private Set<String> matchingPolicies;
    final PKIX.BuilderParams buildParams;
    final X509CertSelector targetCertConstraints;
    static final boolean USE_AIA;
    
    Builder(PKIX.BuilderParams buildParams) {
        this.buildParams = buildParams;
        this.targetCertConstraints = (X509CertSelector)buildParams.targetCertConstraints();
    }
    
    abstract Collection<X509Certificate> getMatchingCerts(State p0, List<CertStore> p1) throws CertStoreException, CertificateException, IOException;
    
    abstract void verifyCert(X509Certificate p0, State p1, List<X509Certificate> p2) throws GeneralSecurityException;
    
    abstract boolean isPathCompleted(X509Certificate p0);
    
    abstract void addCertToPath(X509Certificate p0, LinkedList<X509Certificate> p1);
    
    abstract void removeFinalCertFromPath(LinkedList<X509Certificate> p0);
    
    static int distance(GeneralNameInterface generalNameInterface, GeneralNameInterface generalNameInterface2, int n) {
        switch (generalNameInterface.constrains(generalNameInterface2)) {
            case -1: {
                if (Builder.debug != null) {
                    Builder.debug.println("Builder.distance(): Names are different types");
                }
                return n;
            }
            case 3: {
                if (Builder.debug != null) {
                    Builder.debug.println("Builder.distance(): Names are same type but in different subtrees");
                }
                return n;
            }
            case 0: {
                return 0;
            }
            case 2: {
                break;
            }
            case 1: {
                break;
            }
            default: {
                return n;
            }
        }
        return generalNameInterface2.subtreeDepth() - generalNameInterface.subtreeDepth();
    }
    
    static int hops(GeneralNameInterface generalNameInterface, GeneralNameInterface generalNameInterface2, int n) {
        switch (generalNameInterface.constrains(generalNameInterface2)) {
            case -1: {
                if (Builder.debug != null) {
                    Builder.debug.println("Builder.hops(): Names are different types");
                }
                return n;
            }
            case 3: {
                if (generalNameInterface.getType() != 4) {
                    if (Builder.debug != null) {
                        Builder.debug.println("Builder.hops(): hopDistance not implemented for this name type");
                    }
                    return n;
                }
                else {
                    X500Name x500Name = (X500Name)generalNameInterface;
                    X500Name other = (X500Name)generalNameInterface2;
                    X500Name commonAncestor = x500Name.commonAncestor(other);
                    if (commonAncestor == null) {
                        if (Builder.debug != null) {
                            Builder.debug.println("Builder.hops(): Names are in different namespaces");
                        }
                        return n;
                    }
                    else {
                        return x500Name.subtreeDepth() + other.subtreeDepth() - 2 * commonAncestor.subtreeDepth();
                    }
                }
            }
            case 0: {
                return 0;
            }
            case 2: {
                return generalNameInterface2.subtreeDepth() - generalNameInterface.subtreeDepth();
            }
            case 1: {
                return generalNameInterface2.subtreeDepth() - generalNameInterface.subtreeDepth();
            }
            default: {
                return n;
            }
        }
    }
    
    static int targetDistance(NameConstraintsExtension obj, X509Certificate x509Certificate, GeneralNameInterface name) throws IOException {
        if (obj != null && !obj.verify(x509Certificate)) {
            throw new IOException("certificate does not satisfy existing name constraints");
        }
        X509CertImpl impl;
        try {
            impl = X509CertImpl.toImpl(x509Certificate);
        }
        catch (CertificateException cause) {
            throw new IOException("Invalid certificate", cause);
        }
        if (X500Name.asX500Name(impl.getSubjectX500Principal()).equals(name)) {
            return 0;
        }
        SubjectAlternativeNameExtension subjectAlternativeNameExtension = impl.getSubjectAlternativeNameExtension();
        if (subjectAlternativeNameExtension != null) {
            GeneralNames value = subjectAlternativeNameExtension.get("subject_name");
            if (value != null) {
                int i = 0;
                while (i < value.size()) {
                    if (value.get(i).getName().equals(name)) {
                        return 0;
                    }
                    ++i;
                }
            }
        }
        NameConstraintsExtension nameConstraintsExtension = impl.getNameConstraintsExtension();
        if (nameConstraintsExtension == null) {
            return -1;
        }
        if (obj != null) {
            obj.merge(nameConstraintsExtension);
        }
        else {
            obj = (NameConstraintsExtension)nameConstraintsExtension.clone();
        }
        if (Builder.debug != null) {
            Builder.debug.println("Builder.targetDistance() merged constraints: " + String.valueOf(obj));
        }
        GeneralSubtrees value2 = obj.get("permitted_subtrees");
        GeneralSubtrees value3 = obj.get("excluded_subtrees");
        if (value2 != null) {
            value2.reduce(value3);
        }
        if (Builder.debug != null) {
            Builder.debug.println("Builder.targetDistance() reduced constraints: " + value2);
        }
        if (!obj.verify(name)) {
            throw new IOException("New certificate not allowed to sign certificate for target");
        }
        if (value2 == null) {
            return -1;
        }
        int j = 0;
        while (j < value2.size()) {
            int distance = distance(value2.get(j).getName().getName(), name, -1);
            if (distance >= 0) {
                return distance + 1;
            }
            ++j;
        }
        return -1;
    }
    
    Set<String> getMatchingPolicies() {
        if (this.matchingPolicies != null) {
            Set<String> initialPolicies = this.buildParams.initialPolicies();
            if (!initialPolicies.isEmpty() && !initialPolicies.contains("2.5.29.32.0") && this.buildParams.policyMappingInhibited()) {
                (this.matchingPolicies = new HashSet(initialPolicies)).add("2.5.29.32.0");
            }
            else {
                this.matchingPolicies = Collections.emptySet();
            }
        }
        return this.matchingPolicies;
    }
    
    boolean addMatchingCerts(X509CertSelector selector, Collection<CertStore> collection, Collection<X509Certificate> collection2, boolean b) {
        X509Certificate certificate = selector.getCertificate();
        if (certificate != null) {
            if (selector.match(certificate) && !X509CertImpl.isSelfSigned(certificate, this.buildParams.sigProvider())) {
                if (Builder.debug != null) {
                    Builder.debug.println("Builder.addMatchingCerts: adding target cert\n  SN: " + Debug.toHexString(certificate.getSerialNumber()) + "\n  Subject: " + certificate.getSubjectX500Principal() + "\n  Issuer: " + certificate.getIssuerX500Principal());
                }
                return collection2.add(certificate);
            }
            else {
                return false;
            }
        }
        else {
            boolean n = false;
            for (CertStore certStore : collection) {
                try {
                    for (Certificate certificate2 : certStore.getCertificates(selector)) {
                        if (!X509CertImpl.isSelfSigned((X509Certificate)certificate2, this.buildParams.sigProvider()) && collection2.add((X509Certificate)certificate2)) {
                            n = true;
                        }
                        else {
                            continue;
                        }
                    }
                    if (!b && n) {
                        return true;
                    }
                    continue;
                }
                catch (CertStoreException obj) {
                    if (Builder.debug != null) {
                        Builder.debug.println("Builder.addMatchingCerts, non-fatal exception retrieving certs: " + obj);
                        obj.printStackTrace();
                    }
                    else {
                        continue;
                    }
                }
            }
            return n;
        }
    }
    
    static {
        debug = Debug.getInstance("certpath");
        USE_AIA = (boolean)AccessController.doPrivileged(new GetBooleanAction("com.sun.security.enableAIAcaIssuers"));
    }
}
